import sys
import pandas as pd

from abm1559.utils import rng

class Block:
    """
    An abstract block representation.
    """

    def __init__(self, txs, parent_hash, height):
        self.block_hash = rng.bytes(8)
        self.txs = txs
        self.parent_hash = parent_hash
        self.height = height

    def __str__(self):
        return "Block:\n" + "\n".join([tx.__str__() for tx in self.txs])

    def average_waiting_time(self):
        return 0 if len(self.txs) == 0 else sum([self.height - tx.start_block for tx in self.txs]) / len(self.txs)

class Block1559(Block):
    """
    Blocks filled up with 1559 transactions (see :py:class:`abm1559.chain.Tx1559`).
    """

    def __init__(self, txs, parent_hash, height, basefee):
        super().__init__(txs, parent_hash, height)
        self.basefee = basefee

    def average_tip(self): # in Gwei
        return 0 if len(self.txs) == 0 else sum([tx.tip({
            "basefee": self.basefee,
            "current_block": self.height,
        }) for tx in self.txs]) / len(self.txs) / (10 ** 9)

    def average_gas_price(self): # in Gwei
        return 0 if len(self.txs) == 0 else sum([tx.gas_price({
            "basefee": self.basefee,
            "current_block": self.height,
        }) for tx in self.txs]) / len(self.txs) / (10 ** 9)

    def txs_data(self):
        txs_data = []
        for tx_index, tx in enumerate(self.txs):
            txs_data += [{
                "block_height": self.height,
                "tx_index": tx_index,
                "basefee": self.basefee / (10 ** 9),
                **tx.tx_data(),
            }]
        return txs_data

class Chain:
    """
    A container for :py:class:`abm1559.chain.Block` .
    """

    def __init__(self):
        self.blocks = {}
        self.current_head = (0).to_bytes(8, sys.byteorder)

    def add_block(self, block):
        self.blocks[block.block_hash] = block
        self.current_head = block.block_hash

    def export(self):
        df = []
        for block in self.blocks.values():
            df += block.txs_data()
        return pd.DataFrame(df)



from abm1559.utils import (
    rng,
    constants,
)

class Transaction:
    """
    An abstract superclass for transactions.
    """

    def __init__(self, sender, params, gas_used = constants["SIMPLE_TRANSACTION_GAS"]):
        self.sender = sender
        self.start_block = params["start_block"]
        self.gas_used = gas_used
        self.tx_hash = rng.bytes(8)

    def tx_data(self):
        return {
            "start_block": self.start_block,
            "sender": self.sender,
            "gas_used": self.gas_used,
            "tx_hash": self.tx_hash,
        }


class Tx1559(Transaction):
    """
    Inherits from :py:class:`abm1559.txs.Transaction`. A 1559-type transaction.
    """

    def __init__(self, sender, params, gas_used = constants["SIMPLE_TRANSACTION_GAS"]):
        super().__init__(sender, params, gas_used = gas_used)

        self.gas_premium = params["gas_premium"]
        self.max_fee = params["max_fee"]

    def __str__(self):
        return f"1559 Transaction {self.tx_hash.hex()}: max_fee {self.max_fee}, gas_premium {self.gas_premium}"

    def is_valid(self, params):
        basefee = params["basefee"]
        return self.max_fee >= basefee

    def gas_price(self, params):
        # What the user pays
        basefee = params["basefee"]
        return min(self.max_fee, basefee + self.gas_premium)

    def tip(self, params):
        # What the miner gets
        basefee = params["basefee"]
        return self.gas_price(params) - basefee

    def tx_data(self):
        return {
            **super().tx_data(),
            "gas_premium": self.gas_premium / (10 ** 9),
            "max_fee": self.max_fee / (10 ** 9),
        }

class TxEscalator(Transaction):
    """
    Inherits from :py:class:`abm1559.txs.Transaction`. An escalator-type transaction.
    """

    def __init__(self, sender, params, gas_used = constants["SIMPLE_TRANSACTION_GAS"]):
        super().__init__(sender, params, gas_used = gas_used)

        self.max_block = params["max_block"]
        self.start_premium = params["start_premium"]
        self.max_premium = params["max_premium"]

    def __str__(self):
        return f"Escalator Transaction {self.tx_hash.hex()}: start block {self.start_block}, " + \
                f"max block {self.max_block}, start premium {self.start_premium}, max premium {self.max_premium}"

    def is_valid(self, params):
        current_block = params["current_block"]
        return self.start_block <= current_block and current_block <= self.max_block

    def gas_price(self, params):
        # What the user pays
        current_block = params["current_block"]
        fraction_elapsed = (current_block - self.start_block) / (self.max_block - self.start_block)
        return self.start_premium + fraction_elapsed * (self.max_premium - self.start_premium)

    def tip(self, params):
        # What the miner gets
        # In the escalator, miner gets the whole gas_price
        return self.gas_price(params)



from typing import Dict

import pandas as pd

from abm1559.utils import (
    get_basefee_bounds,
    rng
)

from abm1559.txs import Tx1559

class User:
    """
    Users submit transactions. They have a (randomly chosen) value per Gwei $v$, (we choose per Gwei such that all evaluations of their welfare can be done independently of how much gas the transaction uses).

    The user evaluates the total cost of the transaction in one of two ways, embodied by two different subclasses:

    - :py:class:`abm1559.users.AffineUser`: Incurs a fixed (but randomly selected) cost :math:`c` per unit of time (block-to-block), so :math:`cost(t) = v - c * t`.
    - :py:class:`abm1559.users.DiscountUser`: Incurs a discount :math:`\delta` over time, so :math:`cost(t) = v * (1 - \delta)^t`.

    `AffineUser` or `DiscountUser` are subclassed to create users who send different types of transactions, e.g., 1559-type transactions, escalators or something different. The subclasses should implement:

    - (Optional) `expected_time(params)`: How the user estimates how long they will wait for their transaction to be included.
    - (Optional) `decide_parameters(params)`: Based on their type and `params` (typically, current basefee, length of the queue or salient statistics e.g., distribution of tips in the queue), return transaction parameters.
    - (Requested) `transact(params)`: Queried by the simulation when user is spawned. Returns either a transaction or `None` if they balk.
    """

    def __init__(self, wakeup_block):
        self.pub_key = rng.bytes(8)
        self.wakeup_block = wakeup_block

        # Users have a value / Gwei for the transaction
        self.value = int(rng.uniform(low = 0, high = 20) * (10 ** 9))

    def payoff(self, params):
        gas_price = params["gas_price"]
        return self.cost(params) - gas_price

    def export(self):
        return {
            "pub_key": self.pub_key,
            "value": self.value,
        }

class AffineUser(User):
    """
    Affine users incur a fixed cost per unit of time.
    """

    def __init__(self, wakeup_block):
        super().__init__(wakeup_block)
        self.cost_per_unit = int(rng.uniform(low = 0, high = 1) * (10 ** 9))

    def __str__(self):
        return f"Affine User with value {self.value} and cost {self.cost_per_unit}"

    def cost(self, params):
        current_block = params["current_block"]
        elapsed_time = current_block - self.wakeup_block
        return self.value - self.cost_per_unit * elapsed_time

    def export(self):
        return {
            "user_type": "affine_user",
            "cost_per_unit": self.cost_per_unit,
            **super().export(),
        }

class DiscountUser(User):
    """
    The value of discount users is reduced over time.
    """

    def __init__(self, wakeup_block):
        super().__init__(wakeup_block)
        self.discount_rate = 0.01

    def __str__(self):
        return f"Discount User with value {self.value} and discount rate {self.discount_rate}"

    def cost(self, params):
        current_block = params["current_block"]
        elapsed_time = current_block - self.wakeup_block
        return self.value * (1 - self.discount_rate) ** elapsed_time

    def export(self):
        return {
            "user_type": "discount_user",
            "discount_rate": self.discount_rate,
            **super().export(),
        }

class User1559(AffineUser):
    """
    An affine user sending 1559 transactions.
    """
    # Expects to be included within 5 blocks
    # Prefers not to participate if its expected payoff is negative
    # Computes expected payoff by estimating the worst possible basefee 5 blocks from now
    # Fixed gas_premium

    def expected_time(self, params):
        return 5

    def decide_parameters(self):
        gas_premium = 1 * (10 ** 9)
        # max_fee = self.value - self.expected_time(params = {}) * self.cost_per_unit
        max_fee = self.value
        return {
            "max_fee": max_fee,
            "gas_premium": gas_premium,
            "start_block": self.wakeup_block,
        }

    def worst_expected_basefee(self, current_basefee):
        basefee_bounds = get_basefee_bounds(current_basefee, self.expected_time(params = {}))
        return basefee_bounds["ub"]

    def transact(self, params):
        basefee = params["basefee"]

        tx_params = self.decide_parameters()

        # expected_gas_price = min(self.worst_expected_basefee(basefee) + tx_params["gas_premium"], tx_params["max_fee"])
        expected_gas_price = min(basefee + tx_params["gas_premium"], tx_params["max_fee"])
        expected_block = self.wakeup_block + self.expected_time(params = {})
        expected_payoff = self.payoff({
            "gas_price": expected_gas_price,
            "current_block": expected_block,
        })

        if expected_payoff < 0 or tx_params["max_fee"] < 0:
            return None

        tx = Tx1559(
            sender = self.pub_key,
            params = tx_params,
        )
        return tx

class UserPool:

    def __init__(self):
        self.users = {}

    def add_users(self, users):
        for user in users:
            self.users[user.pub_key] = user

    def get_user(self, pub_key):
        return self.users[user.pub_key]

    def export(self):
        df = []
        for user in self.users.values():
            df += [user.export()]
        return pd.DataFrame(df)



from typing import Sequence, Dict

from abm1559.utils import (
    rng,
    constants,
)

from abm1559.txs import Transaction
from abm1559.chain import Block
from abm1559.users import User, User1559

def spawn_poisson_demand(timestep: int, demand_lambda: float, UserClass) -> Sequence[User]:
    """
    Args:
        timestep (int): Current round
        demand_lambda (float): Rate of arrival, the :math:`lambda` parameter of a Poisson distribution

    Returns:
        Sequence[User1559]: An array of 1559 users
    """

    real = rng.poisson(demand_lambda)
    new_users = [UserClass(timestep) for i in range(real)]
    return new_users

def decide_transactions(demand: Sequence[User], params: Dict) -> Sequence[Transaction]:
    """
    Queries all users and checks who wants to send transactions and who wants to balk.

    Args:
        demand (Sequence[User]): Current demand
        params (Dict): Current simulation environment parameters (e.g., basefee)

    Returns:
        Sequence[Transaction]: An array of transactions
    """

    txs = []
    basefee = params["basefee"]

    for user in demand:
        tx = user.transact({
            "basefee": basefee
        })
        if not tx is None:
            txs.append(tx)

    return txs

def update_basefee(block: Block, basefee: int) -> int:
    """
    Basefee update rule

    Args:
        block (Block): The previous block
        basefee (int): The current basefee

    Returns:
        int: The new basefee
    """

    gas_used = sum([tx.gas_used for tx in block.txs])
    delta = gas_used - constants["TARGET_GAS_USED"]
    return basefee + basefee * delta // constants["TARGET_GAS_USED"] // constants["BASEFEE_MAX_CHANGE_DENOMINATOR"]



from typing import Sequence

from abm1559.txs import Transaction

from abm1559.utils import (
    rng,
    constants,
)

class TxPool:
    """
    Represents the transaction queue.
    """

    def __init__(self):
        self.txs = {}
        self.pool_length = 0

    def add_txs(self, txs: Sequence[Transaction]) -> None:
        """
        Adds `txs` to the queue.

        Args:
            txs (Sequence[Transaction]): The transactions to add

        Returns:
            None
        """

        for tx in txs:
            self.txs[tx.tx_hash] = tx
        self.pool_length += len(txs)

    def remove_txs(self, tx_hashes: Sequence[str]):
        """
        Removes transactions from the queue, indexed by `tx_hashes`.

        Args:
            txs (Sequence[Transaction]): The transactions to add

        Returns:
            None
        """

        for tx_hash in tx_hashes:
            del(self.txs[tx_hash])
        self.pool_length -= len(tx_hashes)

    def average_tip(self, params): # in Gwei
        if self.pool_length == 0:
            return 0
        else:
            return sum([tx.tip(params) for tx in self.txs.values()]) / self.pool_length / (10 ** 9)

    def average_gas_price(self, params):
        if self.pool_length == 0:
            return 0
        else:
            return sum([tx.gas_price(params) for tx in self.txs.values()]) / self.pool_length / (10 ** 9)

    def select_transactions(self, params):
        # Miner side
        basefee = params["basefee"]
        max_tx_in_block = int(constants["MAX_GAS_EIP1559"] / constants["SIMPLE_TRANSACTION_GAS"])

        sorted_valid_demand = sorted(
            [tx for tx in self.txs.values() if tx.is_valid({ "basefee": basefee })],
            key = lambda tx: -tx.tip({ "basefee": basefee })
        )
        selected_txs = sorted_valid_demand[0:max_tx_in_block]

        return selected_txs

    def average_waiting_time(self, current_height):
        return 0 if len(self.txs) == 0 else sum([current_height - tx.start_block for tx in self.txs.values()]) / len(self.txs)

    def __str__(self):
        return "\n".join([tx.__str__() for tx in self.txs.values()])



import numpy as np

constants = {
    "BASEFEE_MAX_CHANGE_DENOMINATOR": 8,
    "TARGET_GAS_USED": 10000000,
    "MAX_GAS_EIP1559": 20000000,
    "EIP1559_DECAY_RANGE": 800000,
    "EIP1559_GAS_INCREMENT_AMOUNT": 10,
    "INITIAL_BASEFEE": 1 * (10 ** 9),
    "PER_TX_GASLIMIT": 8000000,
    "SIMPLE_TRANSACTION_GAS": 21000,
}

rng = np.random.default_rng()

def get_basefee_bounds(basefee, blocks):
    # We want to know how high/low the basefee can be after `blocks` steps, starting from `basefee`
    lb = basefee * (1 - 1.0 / constants["BASEFEE_MAX_CHANGE_DENOMINATOR"])
    ub = basefee * (1 + 1.0 / constants["BASEFEE_MAX_CHANGE_DENOMINATOR"])
    return { "lb": lb, "ub": ub }



from setuptools import setup, find_packages

setup(
    name='abm1559',
    url='https://github.com/barnabemonnot/abm1559',
    author='Barnabé Monnot',
    author_email='barnabe.monnot@ethereum.org',
    packages=find_packages(include=['abm1559', 'abm1559.*']),
    install_requires=['numpy', 'pandas'],
    version='0.0.1',
    license='MIT',
    description='Agent-based simulation environment for EIP 1559',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    python_requires='>=3.8',
)
